/********************************************************************** 
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "game.h"
#include "civmap.h"

#include "gamelog.h"

int gamelog_level;		/* also accessed from stdinhand.c */
bool gamelog_active;
FSSpec gamelog_file;

int secompare1(const void *a, const void *b);

/**************************************************************************
filename can be NULL, which means no logging.
***************************************************************************/
void gamelog_init(FSSpec filename, bool gameloging)
{
  gamelog_level=GAMELOG_FULL;
  gamelog_active=gameloging;
  gamelog_file=filename;
}

/**************************************************************************/
void gamelog_set_level(int level)
{
    gamelog_level=level;
}

/**************************************************************************/
void gamelog(int level, char *message, ...)
{
  short fs;
  OSErr err;
  long count=strlen(message);
  char * temp;
  
  if (level > gamelog_level || !gamelog_active)
    return;
  
  if (level==GAMELOG_MAP){
    strcat(message, '\n');
  } else if (level==GAMELOG_EOT){
  	strcpy(temp, message);	// move message into temp
  	strcpy(message, "*** ");// put *** into message
    strcat(message,temp);	// put temp after message (in reality, put message after ***)
    strcat(message, '\n');	// add a trailing \n
  } else {
  	strcpy(temp, message);	// move message into temp
  	strcpy(message, game.year);// put game.year into message
  	strcat(message, ' ');	// add a aspce after message (game.year)
    strcat(message,temp);	// put temp after message (in reality, put message after game.year)
    strcat(message, '\n');	// add a trailing \n
  }

  err=FSpOpenDF(&gamelog_file, fsWrPerm, &fs);
  if(err != 0)
  {
    fprintf(stderr, "couldn't open game logfile: %s\n", gamelog_file.name);
	exit(1);
  }
    
  err=SetFPos(fs, fsFromLEOF, 0);
  if(err != 0)
  {
    fprintf(stderr, "couldn't move game logfile (%s) marker position for appending.\n", gamelog_file.name);
	exit(1);
  }
    
  err=FSWrite(fs, &count, message);
  if(err != 0)
  {
    fprintf(stderr, "error writing game logfile (%s)\n", gamelog_file.name);
	exit(1);
  }    
  if(count != strlen(message))
  {
    fprintf(stderr, "incomplete write to game logfile (%s)\n", gamelog_file.name);
	exit(1);
  }
}


void gamelog_map(void)
{
  int x,y;
  char* hline;
  if ((hline = (char*)calloc(map.xsize,sizeof(char)))!=0){
    for (y=0;y<map.ysize;y++) {

      for (x=0;x<map.xsize;x++) {
	if (map_get_terrain(x,y)!=T_OCEAN)
	  hline[x]='.';
	else
	  hline[x] = ' ';
      }
      gamelog(GAMELOG_MAP,hline);
    }
  }
}

/*Stuff for gamelog_save()*/
struct player_score_entry {
  int idx;
  int value;
};

int secompare1(const void *a, const void *b)
{
  return (((struct player_score_entry *)b)->value-((struct player_score_entry *)a)->value);
}

void gamelog_save(void){
  /*lifted from historian_largest()*/
  int i;
  char buffer[4096];
  char buf2[4096];
  struct player_score_entry *size=(struct player_score_entry *)malloc(sizeof(struct player_score_entry)*game.nplayers);
  for (i=0;i<game.nplayers;i++) {
    size[i].value=total_player_citizens(&game.players[i]);
    size[i].idx=i;
  }
  qsort(size, game.nplayers, sizeof(struct player_score_entry), (void *)secompare1);
  buffer[0]=0;
  for (i=0;i<game.nplayers;i++) {
    sprintf(buf2,"%2d: %s(%i)  ",i+1, get_race_name_plural(game.players[size[i].idx].race), size[i].value);
    strcat(buffer,buf2);
  }
  gamelog(GAMELOG_EOT,buffer);
  free(size);
}



